#!/bin/sh -u
export LC_ALL=C

readonly THIS_PATH="$(readlink -e "$(dirname "$0")")"

readonly KESL_INCLUDE_SH_PATH="${THIS_PATH}/include"
. "${THIS_PATH}/include-sh"
IncludeSh 'utils'
IncludeSh 'tools'
IncludeSh 'is-product-ready'
IncludeSh 'systemd-unit-path'
IncludeSh 'upgrade-state.inc'
IncludeSh 'offline-event.inc'
IncludeSh 'ini-file.inc'

readonly PRODUCT_NAME='kesl'
readonly PRODUCT_VERSION='11.2.2.5324'
readonly PRODUCT_STARTER='/opt/kaspersky/kesl/libexec/kesl-starter'
readonly PRODUCT_PID='/var/run/kesl.pid'
readonly APP_SETTINGS_FILEPATH='/var/opt/kaspersky/kesl/private/storage/appSettings.xml'
readonly WDSERVER_PID='/var/run/wdserver.pid'
readonly WDSERVER_BIN='/opt/kaspersky/kesl/libexec/wdserver'
readonly WDSERVER_LOG='/var/log/kaspersky/kesl/kesl_wdserver.log'
readonly LAUNCHER_LOG='/var/log/kaspersky/kesl/kesl_launcher.log'
readonly LAUNCHER_FILEPATH='/opt/kaspersky/kesl/libexec/kesl_launcher.sh'
readonly PRODUCT_CONTROL_BIN='/opt/kaspersky/kesl/bin/kesl-control'
readonly VAR_DIRECTORY='/var/opt/kaspersky/kesl/'
readonly SHARED_INIT_DIRECTORY='/opt/kaspersky/kesl/shared/init/'
readonly STORAGE_DIR_NAME='storage/'
readonly VAR_STORAGE_DIRECTORY='/var/opt/kaspersky/kesl/private/storage/'
readonly BASES_INIT_DATA_DATA_DIRECTORY='/opt/kaspersky/kesl/shared/init/updates/data/data/'
readonly BASES_DATA_DIRECTORY='/var/opt/kaspersky/kesl/common/updates/data/data/'
readonly KLNAGENT_INIT_FILENAME='klnagent64'
readonly KLNAGENT_INIT_FILEPATH='/etc/init.d/klnagent64'
readonly INI_FILEPATH='/var/opt/kaspersky/kesl/common/kesl.ini'
readonly MANIFEST_FILEPATH='/opt/kaspersky/kesl/doc/manifest.ini'
readonly MONITOR_OAS_VERSION_FILEPATH='/proc/kav4fs_oas.version'
readonly EVENTS_STORAGE_FILEPATH='/var/opt/kaspersky/kesl/private/storage/events.db'
readonly INSTALL_FILEPATH='/opt/kaspersky/kesl/libexec/install'
readonly INSTALL_LOG_FILEPATH='/var/opt/kaspersky/kesl/private/install.log'
readonly UNINSTALL_FILEPATH='/opt/kaspersky/kesl/libexec/uninstall'
readonly CURRENT='/var/opt/kaspersky/kesl/install-current'
readonly UPGRADE='/var/opt/kaspersky/kesl/install-upgrade'
readonly PREVIOUS='/var/opt/kaspersky/kesl/install-previous'
readonly FAILED='/var/opt/kaspersky/kesl/install-failed'
readonly CURRENT_NAME='install-current'
readonly UPGRADE_NAME='install-upgrade'
readonly PREVIOUS_NAME='install-previous'
readonly FAILED_NAME='install-failed'
readonly INI_FILENAME='kesl.ini'
readonly VAR_COMMON_DIRECTORY='/var/opt/kaspersky/kesl/common/'
readonly INTEGRITY_CHECK_TOOL='/opt/kaspersky/kesl/bin/integrity_check_tool'
readonly INTEGRITY_CHECK_MANIFEST='/opt/kaspersky/kesl/bin/integrity_check.xml'
readonly KLNAGENT_INTEGRITY_CHECK_MANIFEST='/opt/kaspersky/klnagent64/bin/integrity_check.xml'

readonly SHARED_DIRECTORY='/opt/kaspersky/kesl/shared/'
readonly SUPERVISOR_FILENAME='kesl-supervisor'
readonly SUPERVISOR_INIT_D_FILEPATH='/etc/init.d/kesl-supervisor'
readonly SUPERVISOR_INIT_D_SHORT_FILEPATH='/etc/init.d/kesl'
readonly SYSTEMD_SERVICE_FILENAME='kesl-supervisor.service'
readonly SYSTEMD_SERVICE_SHORT_FILENAME='kesl.service'

readonly PREPARE_MODULE_SCRIPT_FILEPATH='/opt/kaspersky/kesl/libexec/prepare_module.sh'
readonly DEFAULT_FILE_MODE_CREATION_MASK='0077'
readonly LOGDATEFMT='+%Y.%m.%d %H:%M:%S.%3N'

readonly LINUX_KERNEL="$(uname -r)"

readonly PRODUCT_START_DELAY=180

readonly WDSERVER_LOG_LEVEL_DEFAULT=0

readonly WDSERVER_SOCK='/var/run/wdserver.sock'

readonly WDSERVER_SHUTDOWN_SIGNAL='15'

readonly WDSERVER_HEADSHOT_SIGNAL='6'

readonly WDRESTART=${KESL_MAX_RESTARTS:-'5'}

readonly WDSERVER_MAX_MEM=${KESL_MAX_MEMORY_USAGE:-'highest/2048MB/50%'}
readonly WDSERVER_MAX_VMEM=${KESL_MAX_VIRTUAL_MEMORY_USAGE:-'off'}
readonly WDSERVER_MAX_SWAP=${KESL_MAX_SWAP_USAGE:-'off'}

readonly WDSERVER_IMMUTABLE_FLAGS="\
--trace-file ${WDSERVER_LOG} \
--n-restart ${WDRESTART} \
--master-timeout ${PRODUCT_START_DELAY} \
--failed-kill NONE \
--wdsock ${WDSERVER_SOCK} \
--max-used-mem ${WDSERVER_MAX_MEM} \
--max-used-vmem ${WDSERVER_MAX_VMEM} \
--max-used-swap ${WDSERVER_MAX_SWAP} \
--shutdown-signal ${WDSERVER_SHUTDOWN_SIGNAL} \
--timeout-after-shutdown-signal 120000 \
--headshot-signal ${WDSERVER_HEADSHOT_SIGNAL} \
--timeout-after-headshot-signal 120000 \
--timeout-after-kill-signal 172800000 \
--umask $(( ${DEFAULT_FILE_MODE_CREATION_MASK} )) \
-p ${WDSERVER_PID} \
-d"

WDSERVER_FLAGS=''

UPGRADE_STATE=''
PATCH_VERSION=''
OLD_VERSION=''

Log()
{
    Echo "$(date -u "${LOGDATEFMT}") lnch  $*"
}

Trace()
{
    Log "|trace| $*"
}

LogAppProcesses()
{
    Log "Print application process state:\n$(PsCommand -ef | grep -F "${PRODUCT_NAME}")"
}

PopulateDataDirectory()
{
    mkdir -p "${BASES_DATA_DIRECTORY}"
    cp -n "${BASES_INIT_DATA_DATA_DIRECTORY}"* "${BASES_DATA_DIRECTORY}"
}

IsFanotifyOn()
{
    local useFanotify=''
    if ! ReadIni "${INI_FILEPATH}" 'General' 'UseFanotify' @useFanotify; then
        return 1
    fi

    if ! echo "${useFanotify}" | grep -iE '^(true|yes|1)$' >/dev/null; then
        return 1
    fi

    return 0
}

IsKernelModuleLoaded()
{
    test -f "${MONITOR_OAS_VERSION_FILEPATH}"
    return $?
}

PrepareKernelModule()
{
    Trace "PrepareKernelModule(): $@"

    local atPath="${1:?}"
    shift

    if ! "${atPath}/${PREPARE_MODULE_SCRIPT_FILEPATH}" "$@"; then
        return 1
    fi

    return 0
}

RemoveKernelModuleFiles()
{
    rm -f "/lib/modules/${LINUX_KERNEL}/misc/kav4fs_oas.ko"
    rm -f "/lib/modules/${LINUX_KERNEL}/misc/redirfs.ko"
}

RevokePatchApproval()
{
    local folder="$1"

    local mnfstIniFile="${folder}/${MANIFEST_FILEPATH}"

    local mnfstPatchId=''
    ReadIni "${mnfstIniFile}" '' 'PatchId' @mnfstPatchId || return $?

    local section='AcceptedAgreement'
    local key='PatchId'

    local patchIds=''
    local patchIds2=''
    if ReadIni "${INI_FILEPATH}" "${section}" "${key}" @patchIds; then
        if [ -n "${patchIds}" ]; then
            DeleteFromList "${patchIds}" ',' "${mnfstPatchId}" @patchIds2
            if [ "${patchIds}" != "${patchIds2}" ]; then
                WriteIni "${INI_FILEPATH}" "${section}" "${key}" "${patchIds2}"
            fi
        fi
    fi
}

CopySymlink()
{
    cp -TP "$1" "$2" || Die "'cp' failed CopySymlink '$1' '$2' (exit code=$?)"
}

MoveSymlink()
{
    mv -fT "$1" "$2" || Die "'mv' failed MoveSymlink '$1' '$2' (exit code=$?)"
}

RemoveSymlink()
{
    rm -f "$1"
}

ReloadKlnagent()
{
    if [ -x "${KLNAGENT_INIT_FILEPATH}" ]; then
        if readlink /proc/1/exe |grep systemd >/dev/null 2>&1; then
            systemctl try-restart "${KLNAGENT_INIT_FILENAME}"
        else
            if "${KLNAGENT_INIT_FILEPATH}" status >/dev/null 2>&1; then
                "${KLNAGENT_INIT_FILEPATH}" restart
            fi
        fi
    fi
}

RunInstUninstAtDir()
{
    local path="$1"
    local script="$2"

    "${path}/${script}" "${path}/${INSTALL_LOG_FILEPATH}"
    return $?
}

Call_GetUpgradeState()
{
    GetUpgradeState @UPGRADE_STATE @PATCH_VERSION @OLD_VERSION
}

SwitchToUpgradeState()
{
    local newState="$1"
    shift

    Log "Switch upgrade state: '${UPGRADE_STATE}' -> '${newState}': $@"

    "$@"

    Call_GetUpgradeState
    if [ "${UPGRADE_STATE}" != "${newState}" ]; then
        Log "WARNING: GetUpgradeState() returned unexpected upgrade state: ${UPGRADE_STATE}; expected: ${newState}"
    fi
}

ExecLauncher()
{
    local ver="$1"

    Trace "ExecLauncher($1)"

    exec "${CURRENT}/${LAUNCHER_FILEPATH}" || Die "Failed to exec ${ver} launcher (exit code=$?)"
}

DoUninstall()
{
    Trace "DoUninstall()"

    RunInstUninstAtDir "${CURRENT}" "${UNINSTALL_FILEPATH}"
}

DoUpgrade()
{
    Trace "DoUpgrade()"

    if ! IsFanotifyOn; then
        RemoveKernelModuleFiles
    fi

    RunInstUninstAtDir "${UPGRADE}" "${INSTALL_FILEPATH}" || return $?

    if [ -f "${PREVIOUS}${EVENTS_STORAGE_FILEPATH}" ]; then
        cp -fpv "${PREVIOUS}${EVENTS_STORAGE_FILEPATH}" "${UPGRADE}${EVENTS_STORAGE_FILEPATH}" || return $?
    fi

    ReloadKlnagent

    return 0
}

DoStart()
{
    Trace "DoStart()"

    Log "Try to start version '${PRODUCT_VERSION}' at install directory '$(readlink ${CURRENT})'"
}

DoUpgradeFailed()
{
    Trace "DoUpgradeFailed()"

    RunInstUninstAtDir "${FAILED}" "${UNINSTALL_FILEPATH}"
}

DoStartFailed()
{
    Trace "DoStartFailed()"

    RunInstUninstAtDir "${FAILED}" "${UNINSTALL_FILEPATH}"
}

DoRollingBack()
{
    Trace "DoRollingBack()"

    Log "Revoke patch approval at rolling back patch '${PATCH_VERSION}'"
    RevokePatchApproval "${FAILED}"

    if ! IsFanotifyOn; then
        RemoveKernelModuleFiles
    fi

    if ! RunInstUninstAtDir "${CURRENT}" "${INSTALL_FILEPATH}"; then
        EmitOfflineEvent 'PatchRollbackError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' ''
    fi

    ReloadKlnagent
}

UpgradeService()
{
    Trace "UpgradeService(): OLD_VERSION=${OLD_VERSION}, PATCH_VERSION=${PATCH_VERSION}"

    local systemdUnitPath=''
    GetSystemdUnitPath @systemdUnitPath
    if [ -n "${systemdUnitPath}" ]; then
        cp -LTfv "${CURRENT}${SHARED_DIRECTORY}${SYSTEMD_SERVICE_FILENAME}" "${systemdUnitPath}/${SYSTEMD_SERVICE_FILENAME}"
        systemctl daemon-reload
    fi
}

AssumeInstalled()
{
    Trace "AssumeInstalled()"

    local cmp=''
    if CompareVersions "${OLD_VERSION}" "${PATCH_VERSION}" @cmp && [ "${cmp}" = '>' ]; then
        Log "Revoke previously installed patch approval at downgrading from '${OLD_VERSION}' to '${PATCH_VERSION}'"
        RevokePatchApproval "${PREVIOUS}"
    fi

    local prevDirName="$(readlink "${PREVIOUS}")"
    SwitchToUpgradeState 'Installed' RemoveSymlink "${PREVIOUS}"

    UpgradeService

    Log "Remove previous version of files at '${VAR_DIRECTORY}${prevDirName}'"
    rm -rf "${VAR_DIRECTORY}${prevDirName}"

    StartIntegrityChecks
}

ProcessUpgradeState()
{
    case "${UPGRADE_STATE}" in
        'Installed')
            return 1
            ;;
        'Uninstall')
            if ! IsFanotifyOn; then
                if IsKernelModuleLoaded; then
                    Log "Cannot do 'Uninstall' now: UseFanotify=false and kernel module is loaded: operating system restart is required"
                    return 1
                fi
                if ! PrepareKernelModule "${UPGRADE}" '--build-only'; then
                    Log "Failed to build new kernel module"
                    SwitchToUpgradeState 'RollingBack' MoveSymlink "${UPGRADE}" "${FAILED}"
                    EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Failed to build new kernel module'
                    return 0
                fi
            fi

            DoUninstall
            SwitchToUpgradeState 'Upgrade' CopySymlink "${CURRENT}" "${PREVIOUS}"
            ;;
        'Upgrade')
            if DoUpgrade; then
                SwitchToUpgradeState 'Start' MoveSymlink "${UPGRADE}" "${CURRENT}"
                ExecLauncher 'new'
            else
                SwitchToUpgradeState 'UpgradeFailed' MoveSymlink "${UPGRADE}" "${FAILED}"
            fi
            ;;
        'Start')
            DoStart
            return 1
            ;;
        'UpgradeFailed')
            DoUpgradeFailed
            SwitchToUpgradeState 'RollingBack' RemoveSymlink "${PREVIOUS}"
            EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Upgrade procedure failed'
            ;;
        'StartFailed')
            DoStartFailed
            SwitchToUpgradeState 'RollingBack' MoveSymlink "${PREVIOUS}" "${CURRENT}"
            EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Failed to start application'
            ExecLauncher 'old'
            ;;
        'RollingBack')
            DoRollingBack
            SwitchToUpgradeState 'RolledBack' CopySymlink "${FAILED}" "${UPGRADE}"
            EmitOfflineEvent 'PatchRollbackCompleted' 'Informational' 'PatchVersion' "${PATCH_VERSION}"
            StartIntegrityChecks
            ;;
        *)
            Log "No actions required for upgrade state '${UPGRADE_STATE}'"
            return 1
            ;;
    esac

    return 0
}

StartKeslIntegrityCheck()
{
    local res=true
    "${INTEGRITY_CHECK_TOOL}" -v -m "${INTEGRITY_CHECK_MANIFEST}" || res=false

    if ! ${res}; then
        EmitOfflineEvent 'IntegrityCheckFailed' 'Error' 'Reason' 'Kesl integrity check failed'
    fi

    return 0
}

StartKlnagentIntegrityCheck()
{
    if [ ! -f "${SC_KLNAGENT_BIN}" ]; then
        return 0
    fi

    local res=true
    "${INTEGRITY_CHECK_TOOL}" -v -m "${KLNAGENT_INTEGRITY_CHECK_MANIFEST}" || res=false
    if ! ${res}; then
        EmitOfflineEvent 'IntegrityCheckFailed' 'Error' 'Reason' 'klnagent integrity check failed'
    fi

    return 0
}

StartIntegrityChecks()
{
    Trace "StartIntegrityChecks()"

    StartKeslIntegrityCheck &
    StartKlnagentIntegrityCheck &
}


LaunchKeslViaWdserver()
{

    "${WDSERVER_BIN}" ${WDSERVER_FLAGS} -e '/usr/sbin/execaps' -c 0x300 -- "${PRODUCT_STARTER}"
    return $?
}


ComposeWdServerFlags()
{
    Trace "ComposeWdServerFlags()"

    local traceLevel="${WDSERVER_LOG_LEVEL_DEFAULT}"

    if [ -f "${APP_SETTINGS_FILEPATH}" ]; then
        local appSettings="$(cat "${APP_SETTINGS_FILEPATH}")"

        local level=''
        RemoveLargestPrefix appSettings '*<TraceLevel>'  @level
        RemoveLargestSuffix level       '</TraceLevel>*' @level

        case "${level}" in
            'None')
                traceLevel=0 ;;
            'NotDetailed')
                traceLevel=300 ;;
            'Detailed')
                traceLevel=600 ;;
            *)
                Log "Unknown 'TraceLevel' value: '${level}'" ;;
        esac
    else
        Log "File '${APP_SETTINGS_FILEPATH}' does not exist: use default value for '--trace-level' option: '${traceLevel}'"
    fi

    WDSERVER_FLAGS="--trace-level ${traceLevel} ${WDSERVER_IMMUTABLE_FLAGS}"
}

RunWdServer()
{
    {
        rm -f "${PRODUCT_PID}"
        rm -f "${WDSERVER_PID}"
    } 2>/dev/null 1>&2

    ComposeWdServerFlags
    Log "Running wdserver with options:\n${WDSERVER_FLAGS}"
    LaunchKeslViaWdserver
    local wdRes=$?
    if [ "${wdRes}" -eq 0 ]; then
        Log "wdserver runnning!"
        if [ "${UPGRADE_STATE}" = 'Start' ]; then
            AssumeInstalled

            {
                local prevMajor=''
                local prevMinor=''
                ParseVersionString "${OLD_VERSION}" @prevMajor @prevMinor ||
                    Die "error on parse previous version: '${OLD_VERSION}'"

                if [ "${prevMajor}" = '11' ] && [ "${prevMinor}" = '1' ]; then
                    local svPid=${PPID}
                    while kill -0 ${svPid} >/dev/null 2>&1; do
                        Log "Wait for the old 'supervisor' to die (sleep 1)..."
                        sleep 1
                    done
                fi
            }
        fi
        return 0
    fi

    if [ "${UPGRADE_STATE}" = 'Start' ]; then
        LogAppProcesses
        Log "*** Failed to start wdserver *** (wdRes=${wdRes})"
        Log "*** Failed to start application the first time after upgrade ***"
        SwitchToUpgradeState 'StartFailed' CopySymlink "${CURRENT}" "${FAILED}"
        ProcessUpgradeState
    else
        LogAppProcesses
        DieRc ${wdRes} "*** Failed to start wdserver *** (wdRes=${wdRes})"
    fi
}

Main()
{
    Repeat 85 '='
    Trace "Main(): pid=$$, PPID=$PPID"

    while :; do
        Call_GetUpgradeState || Die "Failed to get upgrade state (exit code=$?): returned state='${UPGRADE_STATE}'"
        Log "Current upgrade state: '${UPGRADE_STATE}'"
        if ! ProcessUpgradeState; then
            break
        fi
    done

    case "${UPGRADE_STATE}" in
        ''|'Installed'|'RolledBack')
            Log "Start Integrity checks..."
            StartIntegrityChecks
            ;;
        *)
            Log "Skip integrity checks (Upgrade state = ${UPGRADE_STATE})..."
            ;;
    esac

    PopulateDataDirectory

    RunWdServer

    Log "${PRODUCT_NAME} successfully started."
    return 0
}

umask "${DEFAULT_FILE_MODE_CREATION_MASK}" && Main "$@" >>"${LAUNCHER_LOG}" 2>&1
