package KL::PlaformSpecificUtils;

use strict;

use base 'KL::AppBase';

use POSIX qw(setlocale);
BEGIN
{
    my $lang = substr(setlocale("LC_ALL"), 0, 2);
    my @LFS_LANGS = split(/\s*,\s*/, "%LFS_LICENSE_LANGS%");
    if (grep {$_ eq $lang} @LFS_LANGS)
    {
        eval "require KL::Localization::$lang; KL::Localization::$lang->import();";
    }
    else
    {
        require KL::Localization::en;
        KL::Localization::en->import();
    }
}

sub new
{
    my $class = shift;
    my $appid = shift;

    my $self = $class->SUPER::new($appid);

    push @{$self->{'COMMANDS'}}, {
        'id' => 'build_oas',
        'spec' => 'b|build:s',
        'sub' => 'do_build_oas_module',
        'default' => undef,
        'descr' => $MSG_CMD_HELP_BUILD,
    };
    
    return $self;
}

sub do_build_oas_module
{
    KL::Utils::output($MSG_CHECKING_KERNEL_MODULE_AVAILABLE);

    my $os_rev = `uname -r`;
    $os_rev =~ s/\n//;

    my $modules_path = "/lib/modules/${os_rev}/misc";

    my $result = (-f "$modules_path/kav4fs_oas.ko");
    if ($result)
    {
        KL::Utils::output($MSG_USING_KERNEL_MODULE_INTERCEPTOR);
        return 1;
    }
    KL::Utils::output($MSG_KERNEL_MODULE_MISSING_FOR_KERNEL_VERSION);
    KL::Utils::warning($MSG_OAS_DISABLED_WARN);
    return 0;
}

1;

